unit WarnAssignToFunctionName;

{ AFS 21 Sept 2001

 warn of assignment to function name in old TurboPascal code

 ie
  function Fred: integer;
  begin
    Fred := 3;
  end;

 should be

  function Fred: integer;
  begin
    Result := 3;
  end;
}

interface

uses
  {  delphi } Classes,
  {  local } TokenType, Token, Warn;

type
  TWarnAssignToFunctionName = class(TWarn)
  private
    // function names - need to cope with nested functions
    fsNameStack: TStringList;

    function TopName: string;
    procedure PopName;
  protected
    function OnProcessToken(const pt: TToken): TToken; override;
  public
    constructor Create; override;
    destructor Destroy; override;
  end;

implementation

uses
    { delphi } SysUtils,
    { local } WordMap;

constructor TWarnAssignToFunctionName.Create;
begin
  inherited;

  fsNameStack := TStringList.Create;
end;

destructor TWarnAssignToFunctionName.Destroy;
begin
  FreeAndNil(fsNameStack);
  inherited;
end;

function TWarnAssignToFunctionName.TopName: string;
begin
  if fsNameStack = nil then
    exit;

  if fsNameStack.Count < 1 then
    exit;

  Result := fsNameStack[fsNameStack.Count - 1];
end;

procedure TWarnAssignToFunctionName.PopName;
begin
  if fsNameStack = nil then
    exit;

  if fsNameStack.Count < 1 then
    exit;

  fsNameStack.Delete(fsNameStack.Count - 1);
end;


function TWarnAssignToFunctionName.OnProcessToken(const pt: TToken): TToken;
var
  ptNext, ptNextTest: TToken;
  lsNewName: string;
begin
  Result := pt;

  if fsNameStack = nil then
    exit;

  {  look for:
    - in procedure header, a function/proc name
    - in following procedure body, that same name
  }
  if pt.ClassDefinitionSection <> cdNotInClassDefinition then
    exit;

  case (pt.ProcedureSection) of
    psProcedureDefinition:
    begin
      if (pt.Word in ProcedureWords) then
      begin
        ptNext := FirstSolidToken;
        // should be a name?
        assert((ptNext.TokenType in TextualTokens), 'Unexpected token: ' + ptNext.Describe);
        lsNewName := ptNext.SourceCode;

        { this could be a member fn, eg. function MyClass.GetSomevalue: integer;
          in which case the name we want is after a dot,
        }
        ptNextTest := NextSolidTokenAfter(ptNext);
        if ptNextTest.TokenType = ttDot then
        begin
          ptNextTest := NextSolidTokenAfter(ptNextTest);
          if ptNext.TokenType = ttWord then
            lsNewName := ptNextTest.SourceCode;
        end;

        fsNameStack.Add(lsNewName);
      end;
    end;

    psProcedureBody:
    begin
      // check if we have exited a contained procedure
      while pt.ProcedureNestingLevel < fsNameStack.Count do
        PopName;


      if ((pt.TokenType = ttWord) and (pt.SourceCode = TopName)) then
      begin
        LogWarning(pt,
          'Assignment to the function name "' + TopName +
          '" is deprecated, Use assignment to "Result"');
      end;
    end;
    psNotInProcedure:
    begin
      fsNameStack.Clear;
    end;
  end;

end;

end.
